/*

    Program: Color tracking program - YUV version 
    Copyright (C) 1999  A. Schiffler

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    To contact the author try these:
    Andreas Schiffler, 
    ZKM, Lorezstr. 19, 76135 Karlsruhe, Germany
    schiffler@zkm.de

*/

/* Color tracking program - YUV version  */

/* Requires Matrox meteor framegrabber card+driver (included). */
/* Requires X-visual at 15bpp (startx -- -bpp 16 -weight 555) */

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <sys/signal.h>
#include <math.h>

#include "ioctl_meteor.h"

extern int errno;

static void setupDisplay();

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/extensions/XShm.h>
#include <sys/shm.h>
#include <X11/keysym.h>

#define NPORT 2002

#define ROWS 480
#define COLS 640

#define SIZE (ROWS * COLS * 2)

#define WINDOW_SIZE_X COLS
#define WINDOW_SIZE_Y ROWS

static Display* display_;
static GC gc_;
static Window wtwin_;
static XEvent event;
static int in_loop;

static XImage *shmimage;
static XShmSegmentInfo* shminfo;

static char *DevName;
static char *mmbuf, *fbuf;

static unsigned char chcv=44;
static unsigned char bright=148; 
static unsigned char chroma=114; 
static unsigned char contrast=74;
static unsigned char hue=5;       

static int the_grabber;
static struct sigaction sigact; 

/* Tracking bounds */

static int boundx1 = 0;
static int boundx2 = COLS-1;
static int boundy1 = 0;
static int boundy2 = ROWS-1;

static int boundmode = 0;	/* 0=resize and 1=move */

/* Toggles */

static int SHOW_FPS=1;
static int SHOW_RESULT=1;

/* Statistics */

static int framecounter=0;
static float LastTime,TheCurrentTime;

/* Image Processing Settings and Results */

#define MAX_TARGETS	10
#define COUNT_THRESHOLD	10
		
static int num_colors = 0;

struct ColorTarget {
	float color_min_y;
	float color_max_y;
	float color_min_u;
	float color_max_u;
	float color_min_v;
	float color_max_v;
	long result_sumx;
	long result_sumy;
	long result_count;
	float result_xs;
	float result_ys;
	float result_x;
	float result_y;
} targets[MAX_TARGETS];
	
/* Function Protptype for Framegrabber Callback */

void gotframe(int x);

/* Multithreading */

#include <pthread.h>
#include <semaphore.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sched.h>

static pthread_t thread;
      
/* Semaphores */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#define SEM_RESOURCE_MAX        1

key_t semkey;
int semset_id;

/* Uptime of computer in seconds */

float GetRelativeTime()
{
 FILE *infile;
 float thetime;
  
 /* Get the uptime for FPS calculations */
 infile=fopen("/proc/uptime","r");
 fscanf(infile,"%f",&thetime);
 fclose(infile);
      
 return(thetime);
}

/* Configuration file */

static char cnffilename[256];

void LoadConfiguration()
{
 FILE *cnffile;
 int i;
 static int ichcv;
 static int ibright; 
 static int ichroma; 
 static int icontrast; 
 static int ihue; 

 cnffile=fopen(cnffilename,"r");
 if (cnffile != NULL) {
  fscanf(cnffile,"%i",&SHOW_RESULT);
  fscanf(cnffile,"%i",&boundx1);
  fscanf(cnffile,"%i",&boundx2);
  fscanf(cnffile,"%i",&boundy1);
  fscanf(cnffile,"%i",&boundy2);
  fscanf(cnffile,"%i",&ibright);
  fscanf(cnffile,"%i",&ichroma);
  fscanf(cnffile,"%i",&icontrast);
  fscanf(cnffile,"%i",&ihue);
  fscanf(cnffile,"%i",&ichcv);
  bright=ibright;
  chroma=ichroma;
  contrast=icontrast;
  hue=ihue;
  chcv=ichcv;
  fscanf(cnffile,"%i",&num_colors);
  if ((num_colors>0) && (num_colors<MAX_TARGETS)) {
   for (i=0; i<num_colors; i++) {
    fscanf(cnffile,"%f %f %f %f %f %f",
     &targets[i].color_min_y,&targets[i].color_max_y,
     &targets[i].color_min_u,&targets[i].color_max_u,
     &targets[i].color_min_v,&targets[i].color_max_v);
   }
  } 
  fclose(cnffile);

  fprintf (stderr,"Configuration loaded from file: %s\n",cnffilename);
 }
}

void SaveConfiguration()
{
 FILE *cnffile;
 int i;
 static int ichcv;
 static int ibright; 
 static int ichroma; 
 static int icontrast; 
 static int ihue; 
 
 cnffile=fopen(cnffilename,"w");
 if (cnffile != NULL) {
  fprintf(cnffile,"%i\n",SHOW_RESULT);
  fprintf(cnffile,"%i\n",boundx1);
  fprintf(cnffile,"%i\n",boundx2);
  fprintf(cnffile,"%i\n",boundy1);
  fprintf(cnffile,"%i\n",boundy2);
  ibright=bright;
  ichroma=chroma;
  icontrast=contrast;
  ihue=hue;
  ichcv=chcv;
  fprintf(cnffile,"%i\n",ibright);
  fprintf(cnffile,"%i\n",ichroma);
  fprintf(cnffile,"%i\n",icontrast);
  fprintf(cnffile,"%i\n",ihue);
  fprintf(cnffile,"%i\n",ichcv);
  fprintf(cnffile,"%i\n",num_colors);
  if (num_colors>0) {
   for (i=0; i<num_colors; i++) {
    fprintf(cnffile,"%f %f %f %f %f %f\n",
     targets[i].color_min_y,targets[i].color_max_y,
     targets[i].color_min_u,targets[i].color_max_u,
     targets[i].color_min_v,targets[i].color_max_v);
   }
  } 
  fclose(cnffile);
 
  fprintf (stderr,"Configuration saved to file: %s\n",cnffilename);
 }
}

/* SystemV type semaphores */

void createsem(int *sid)
{
 if ((*sid=semget(IPC_PRIVATE, 1, 0666 | IPC_CREAT))== -1) {
  perror("semget: error.");
  exit(0);
 }
 if (semctl(*sid,0,SETVAL,1)==-1) {
  perror("setctl: setval failed.");
 } 
}

void semlock(int *sid)
{
 struct sembuf lockbuf = {0,-1,0};
 
 if(semop(*sid,&lockbuf,1)==-1) {
  perror("semop: error on lock.");
 } 
}

void semunlock(int *sid)
{
 struct sembuf unlockbuf = {0,1,0};
 
 /* Attempt to unlock the semaphore set */
 if(semop(*sid,&unlockbuf,1) == -1) {
  perror("semop: error on unlock");
 } 
}

int semtrywait(int *sid)
{
 int sval;
 
 if ((sval=semctl(*sid,0,GETVAL,NULL))==-1) {
  perror("semctl: error on trywait");
  sval=0;
 } 
 return(sval);
}

void removesem(int *sid)
{
 if (semctl(*sid,0,IPC_RMID,NULL)==-1) {
  perror("semctl: error on remove");
 } 
}

/* Socket includes */

#include <netinet/in.h>

#include <limits.h>
#include <netdb.h>
#include <arpa/inet.h>

#include <sys/socket.h>
#include <errno.h>

/* -------- Converts ascii text to in_addr struct */

struct in_addr *atoaddr(address)
char *address;
{
  struct hostent *host;
  static struct in_addr saddr;

  /* First try it as aaa.bbb.ccc.ddd. */
  saddr.s_addr = inet_addr(address);
  if (saddr.s_addr != -1) {
    return &saddr;
  }
    
  host = gethostbyname(address);
  if (host != NULL) {
    return (struct in_addr *) *host->h_addr_list;
  }
  return NULL;
}

/* ---------- Socket IO -------------------- */
   
int sock_write (int sockfd, char *buf, size_t count)
{
 size_t bytes_sent = 0;
 int this_write;
       
 while (bytes_sent < count) {
  do {
   this_write = write(sockfd, buf, count - bytes_sent);
  } while ( (this_write < 0) && (errno == EINTR) );
  if (this_write <= 0) {
   return(this_write);
  } 
  bytes_sent += this_write;
  buf += this_write;
 }
 return(count);
}

int sock_read (int sockfd, char *buf, size_t count)
{
 size_t bytes_read = 0;
 int this_read;
       
 while (bytes_read < count) {
  do {
   this_read = read(sockfd, buf, count - bytes_read);
  } while ( (this_read < 0) && (errno == EINTR) );
  if (this_read <= 0) {
   return(this_read);
  } 
  bytes_read += this_read;
  buf += this_read;
 }
 return(count);
}
      
int sock_gets (int sockfd, char *str, size_t count)
{
  int bytes_read;
  int total_count = 0;
  char *current_position;
  char last_read = 0;

  current_position = str;
  while ((last_read != 10) && (last_read !=13)) {
    bytes_read = read(sockfd, &last_read, 1);
    if (bytes_read <= 0) {
     return bytes_read; 
    }
    if ( (total_count < count) && (last_read != 10) && (last_read !=13) ) {
      current_position[0] = last_read;
      current_position++;
      total_count++;
    }
  }
  if (count > 0) current_position[0] = 0;
  return total_count;
}


void GlobalExit()
{
 /* Unset loop flag */
 in_loop=0;
 
 /* Stop automatic grabbing */
 sigact.sa_handler=NULL; 
 sigemptyset(&sigact.sa_mask);
 sigact.sa_flags=0;
 if ( sigaction(SIGUSR2, &sigact, NULL) ) {
    perror("sigaction failed");
    exit(0);
 }
 
 /* Clean up semaphore */
 removesem(&semset_id);

 /* Quit program */
 exit(0);
}

void hline (char *framebuffer,int x, int y, int length)
{
 /* Check */
 if (x<0) x=0;
 if (y<0) y=0;
 if (length<0) length=0;
 if (x>(COLS-1)) x=(COLS-1);
 if (y>(ROWS-1)) y=(ROWS-1);
 if ((x+length)>(COLS-1)) length=(COLS-1)-x;
 /* Draw */
 framebuffer += (2*(x + y*COLS));
 memset (framebuffer,255,length*2);
}

void vline (char *framebuffer,int x, int y, int length)
{
 int i;
 
 /* Check */
 if (x<0) x=0;
 if (y<0) y=0;
 if (length<0) length=0;
 if (x>(COLS-1)) x=(COLS-1);
 if (y>(ROWS-1)) y=(ROWS-1);
 if ((y+length)>(ROWS-1)) length=(ROWS-1)-y;
 /* Draw */
 framebuffer += (2*(x + y*COLS));
 for (i=0; i<length; i++) {
  memset (framebuffer,255,2);
  framebuffer += (2*COLS);
 }
}

SetChcv()
{
 fprintf (stderr,"Chcv set to %i\n",chcv);
 if (ioctl(the_grabber, METEORSCHCV, &chcv) < 0) {
  perror("ioctl set chcv failed");
  GlobalExit();
 }
} 

void SetBrightness()
{
 fprintf (stderr,"Brightness set to %i\n",bright);
 if (ioctl(the_grabber, METEORSBRIG, &bright) < 0) {
  perror("ioctl set bright failed");
  GlobalExit();
 }
}

void SetChroma()
{
 fprintf (stderr,"Chroma set to %i\n",chroma);
 if (ioctl(the_grabber, METEORSCSAT, &chroma) < 0) {
  perror("ioctl set chroma failed");
  GlobalExit();
 }
}

void SetContrast()
{
 fprintf (stderr,"Contrast set to %i\n",contrast);
 if (ioctl(the_grabber, METEORSCONT, &contrast) < 0) {
  perror("ioctl set contrast failed");
  GlobalExit();
 } 
}

void SetHue()
{
 fprintf (stderr,"Hue set to %i\n",hue);
 if (ioctl(the_grabber, METEORSHUE, &hue) < 0) {
  perror("ioctl set hue failed");
  GlobalExit();
 } 
}

#define BLENGTH 10

int range (int x, int y)
{
 if ((x<0) || (x>(COLS-1))) return(0);
 if ((y<0) || (y>(ROWS-1))) return(0);
 return(1);
}

/* Track RGB */

void DefaultSetup()
{
 num_colors=1;
 
 targets[0].color_min_y=0.9;
 targets[0].color_max_y=1.0;
 targets[0].color_min_u=0.0;
 targets[0].color_max_u=0.1;
 targets[0].color_min_v=0.0;
 targets[0].color_max_v=0.1;
}

/* Colormatch-lookup-table */
static int color_lookup[32768];

void InitLookup()
{
 int i,j;
 int rpart;
 int gpart;
 int bpart;
 float r,g,b;
 float y,u,v;
 float ux,vx;
 float dist;
  
 fprintf (stderr, "Initializing lookup table ...\n"); 
 for (i=0; i<32768; i++) {
  /* Mark as unused */
  color_lookup[i]=0;
  /* Extract color components */
  rpart=(((unsigned short)i >> 10) & 15);
  gpart=(((unsigned short)i >> 5) & 15);
  bpart= ((unsigned short)i & 15);
  /* Convert to yuv */
  r=((float)rpart/15.0);
  g=((float)gpart/15.0);
  b=((float)bpart/15.0);
  y=0.299*r+0.587*g+0.114*b;
  u=0.493*(b-y);
  v=0.877*(r-y);
  /* Determine match */
  for (j=0; j<num_colors; j++) {
   if ((y>targets[j].color_min_y) && (y<targets[j].color_max_y) &&
       (u>targets[j].color_min_u) && (u<targets[j].color_max_u) && 
       (v>targets[j].color_min_v) && (v<targets[j].color_max_v)    ) {   
    color_lookup[i]=j+1;
   } 
  }
 }
}

/* Main processing routine statistics */

static int clickcount=0; 
static float ymin,ymax,ysum;
static float umin,umax,usum;
static float vmin,vmax,vsum;

/* 
TODO: The framebuffer access does not use the offset provided by the driver
as is required for some motherboard/meteor combinations. For a reference
see the sample applications of the new meteor driver.
*/

void gotframe(int x)
{
  int i,j;
  char aline[128];
  KeySym key;
  XComposeStatus compose;
  int charcount;
  unsigned short *curpos;
  int match_number;
  struct ColorTarget *match_target;
  float r,g,b,y,u,v;
  int rpart,gpart,bpart;
      
  if (SHOW_FPS) {
   /* Every 100th frame, output FPS number */
   framecounter++;
   if (framecounter==100) {
    TheCurrentTime = GetRelativeTime();
    if (TheCurrentTime>LastTime) {
     fprintf (stderr,"Display Speed = %4.1f FPS\n",100.0/(TheCurrentTime-LastTime)); 
    }
    framecounter=0;
    LastTime=TheCurrentTime;
   }   
  }
                                               
  /* Copy image to screen */
  memcpy(shmimage->data, fbuf, SIZE);
  
  /* Reset sums */
  for (i=0; i<num_colors; i++) {
   targets[i].result_sumx = 0;
   targets[i].result_sumy = 0;
   targets[i].result_count = 0;
  }
  
  /* Work on selected part of image */
  curpos=(unsigned short *)shmimage->data;
  curpos += (boundx1+COLS*boundy1);
  for (i=boundy1; i<boundy2; i++) {
   for (j=boundx1; j<boundx2; j++) {
    match_number=color_lookup[(*curpos & 0x7fff)];
    if (match_number>0) {
     match_target=&targets[match_number-1];
     match_target->result_sumx += j;
     match_target->result_sumy += i;
     match_target->result_count ++;
     if (SHOW_RESULT) {
      *curpos=0xffff;
     } 
    }
    curpos++;
   }
   curpos += (COLS-boundx2+boundx1);
  }
  
  /* Store distances in thread-shared variables */
  semlock(&semset_id);
  for (i=0; i<num_colors; i++) {
   if (targets[i].result_count>COUNT_THRESHOLD) {
    targets[i].result_xs=targets[i].result_sumx/targets[i].result_count;
    targets[i].result_ys=targets[i].result_sumy/targets[i].result_count;
    targets[i].result_x=(targets[i].result_xs-(float)boundx1)/(float)(boundx2-boundx1);
    targets[i].result_y=(targets[i].result_ys-(float)boundy1)/(float)(boundy2-boundy1);
   }
  } 
  semunlock(&semset_id);
  
  /* Show bounds for the tracking */
  /* Upper left */
  hline ((char *)shmimage->data,boundx1,boundy1, BLENGTH);
  vline ((char *)shmimage->data,boundx1,boundy1, BLENGTH);
  /* Lower left */
  hline ((char *)shmimage->data,boundx1,boundy2, BLENGTH);
  vline ((char *)shmimage->data,boundx1,boundy2-BLENGTH, BLENGTH);
  /* Upper right */
  hline ((char *)shmimage->data,boundx2-BLENGTH,boundy1, BLENGTH);
  vline ((char *)shmimage->data,boundx2,boundy1, BLENGTH);
  /* Lower right */
  hline ((char *)shmimage->data,boundx2-BLENGTH,boundy2, BLENGTH);
  vline ((char *)shmimage->data,boundx2,boundy2-BLENGTH, BLENGTH);
  
  /* Show cursor mode */ 
  if (boundmode==0) {
   hline ((char *)shmimage->data,(boundx1+boundx2)/2-BLENGTH,boundy1, 2*BLENGTH);
   hline ((char *)shmimage->data,(boundx1+boundx2)/2-BLENGTH,boundy2, 2*BLENGTH);
   vline ((char *)shmimage->data,boundx1,(boundy1+boundy2)/2-BLENGTH, 2*BLENGTH);
   vline ((char *)shmimage->data,boundx2,(boundy1+boundy2)/2-BLENGTH, 2*BLENGTH);
  } else {
   hline ((char *)shmimage->data,(boundx1+boundx2)/2-BLENGTH,(boundy1+boundy2)/2, 2*BLENGTH);
   vline ((char *)shmimage->data,(boundx1+boundx2)/2,(boundy1+boundy2)/2-BLENGTH, 2*BLENGTH);
  }
 
  /* Show tracking results */
  for (i=0; i<num_colors; i++) {
   hline ((char *)shmimage->data,targets[i].result_xs-BLENGTH,targets[i].result_ys, 2*BLENGTH);
   vline ((char *)shmimage->data,targets[i].result_xs,targets[i].result_ys-BLENGTH, 2*BLENGTH);
  }
     
  /* Check keyboard input */
  if (XPending(display_)>0) {
   XNextEvent(display_, &event);
   switch (event.type) {
    case ButtonPress:
     /* Get color values */
     curpos=(unsigned short *)fbuf;
     curpos += (event.xbutton.x+COLS*event.xbutton.y);
     rpart=((*curpos >> 10) & 15);
     gpart=((*curpos >> 5) & 15);
     bpart=(*curpos & 15);
     /* Show click */
     hline ((char *)shmimage->data,event.xbutton.x-BLENGTH,event.xbutton.y, 2*BLENGTH);
     vline ((char *)shmimage->data,event.xbutton.x,event.xbutton.y-BLENGTH, 2*BLENGTH);
     /* Calculate and print U and V components */
     r=((float)rpart/15.0);
     g=((float)gpart/15.0);
     b=((float)bpart/15.0);
     y=0.299*r+0.587*g+0.114*b;
     u=0.493*(b-y);
     v=0.877*(r-y);
     /* Work out statistice */
     if (clickcount==0) {
      ymin=y; ymax=y;
      umin=u; umax=u;
      vmin=v; vmax=v;
     }
     clickcount++;
     if (y<ymin) ymin=y;
     if (y>ymax) ymax=y;
     ysum += y;
     if (u<umin) umin=u;
     if (u>umax) umax=u;
     usum += u;
     if (v<vmin) vmin=v;
     if (v>vmax) vmax=v;
     vsum += v;
     
     fprintf (stderr,"%i: Y=%5.3f/%5.3f/%5.3f U=%5.3f/%5.3f/%5.3f V=%5.3f/%5.3f/%5.3f (min/avg/max)\n",
       clickcount,
       ymin,ysum/clickcount,ymax,
       umin,usum/clickcount,umax,
       vmin,vsum/clickcount,vmax);
     break;
    case KeyPress:
     key=XLookupKeysym((void *)&event,0);
     switch (key) {
      case XK_Home:
       boundmode += 1;
       boundmode &= 1;
       fprintf (stderr,"Bounding-box-mode set to %i\n",boundmode);
       break;
      case XK_Left:
       if (boundmode==0) {
        if ((boundx1>0) && (boundx2<(COLS-1))) {
         boundx1 -= 1;
         boundx2 += 1;
        } 
       } else {
        if (boundx1>0) {
         boundx1 -= 1;
         boundx2 -= 1;
        } 
       }
       break;
      case XK_Right:
       if (boundmode==0) {
        if ((boundx2-boundx1)>2) {
         boundx1 += 1;
         boundx2 -= 1;
        }
       } else {
        if (boundx2<(COLS-1)) {
         boundx1 += 1;
         boundx2 += 1;
        } 
       }
       break;
      case XK_Up:
       if (boundmode==0) {
        if ((boundy2-boundy1)>2) {
         boundy1 += 1;
         boundy2 -= 1;
        }
       } else {
        if (boundy1>0) {
         boundy1 -= 1;
         boundy2 -= 1;
        } 
       }
       break;
      case XK_Down:
       if (boundmode==0) {
        if ((boundy1>0) && (boundy2<(ROWS-1))) {
         boundy1 -= 1;
         boundy2 += 1;
        } 
       } else {
        if (boundy2<(ROWS-1)) {
         boundy1 += 1;
         boundy2 += 1;
        } 
       }
       break;
      case XK_y:
       if (contrast>0) {
        contrast--;
        SetContrast();
       }  
       break;
      case XK_x:
       if (contrast<255) {
        contrast++;
        SetContrast();
       }
       break; 
      case XK_o:
       if (chcv>0) {
        chcv--;
        SetChcv();
       }  
       break;
      case XK_p:
       if (chcv<255) {
        chcv++;
        SetChcv();
       }
       break; 
      case XK_b:
       if (bright>0) {
        bright--;
        SetBrightness();
       }  
       break;
      case XK_n:
       if (bright<255) {
        bright++;
        SetBrightness();
       }
       break; 
      case XK_c:
       if (chroma>0) {
        chroma--;
        SetChroma();
       }
       break; 
      case XK_v:
       if (chroma<127) {
        chroma++;
        SetChroma();
       }
       break; 
      case XK_d:
       if (hue>0) {
        hue--;
        SetHue();
       }
       break; 
      case XK_f:
       if (hue<127) {
        hue++;
        SetHue();
       }
       break; 
      case XK_0:
       SaveConfiguration();
       break; 
      case XK_q:
       GlobalExit();
       break; 
        
     }
     break;
    case MappingNotify:
     XRefreshKeyboardMapping((void  *)&event);
     break;
   }
  }  

  /* Draw screen onto display */
  XShmPutImage(display_, wtwin_, gc_, shmimage,
	       0, 0, 0, 0, WINDOW_SIZE_X, WINDOW_SIZE_Y, False);
  XSync(display_, 0);
  
}

void NetworkThread()
{
 u_short port;
 struct sockaddr_in address;
 int listening_socket = -1;
 int connected_socket = -1;
 int connected = 0;
 char buffer[128];
 int reuse_addr = 1;
 int length;
 char hostname[128];
 int cindex;
 
 /* Setup internet address information */  
 port=NPORT;
 memset((char *) &address, 0, sizeof(address));
 gethostname(hostname,sizeof(hostname));
 address.sin_family = AF_INET;
 address.sin_port = htons(port);
 address.sin_addr.s_addr = INADDR_ANY;

 /* Create listening socket */
 listening_socket = socket(AF_INET, SOCK_STREAM, 0);
 if (listening_socket < 0) {
  perror("socket");
  GlobalExit();
 }
 if (bind(listening_socket,&address,sizeof(address)) < 0) {
  perror("bind");
  close(listening_socket);
  GlobalExit();
 }
 listen(listening_socket, 1);
 
 /* Get a connection and handle it */
 while (in_loop) {
  fprintf (stderr,"Waiting for connection on port %i.\n",NPORT);
  length=sizeof(address);
  connected_socket = accept(listening_socket, &address, &length);
  if ((connected_socket < 0) && (errno != EINTR) && (errno != 9)) {
   perror("accept");
  } else {
   fprintf (stderr,"Connected ...\n");
   connected=1;
   while (connected) {
    if (sock_gets(connected_socket,buffer,128)>0) {
     if (buffer[0]=='l') {
      connected=0;
     } else {
      cindex=atoi(buffer);
      if ((cindex>0) && (cindex<=num_colors)) {
       cindex--;
       semlock(&semset_id);
       snprintf (buffer,128,"%f %f\n",targets[cindex].result_x,targets[cindex].result_y);
       semunlock(&semset_id);
       connected=(sock_write(connected_socket,buffer,strlen(buffer))>0);
      }
     } 
    }   
   }
   close(connected_socket);
  }
 }
 close(listening_socket);
 GlobalExit();
}

void main(int argc, char **argv)
{

  struct meteor_geomet geo;
  struct meteor_frame_offset off;
  static char DefaultName[]="/dev/mmetfgrab0";
  int c,i,modeflag,inputflag;

  /* Title */ 
  fprintf (stderr,"%s: Color Tracking for Matrox Meteor FG and Socket Interface\nA. Schiffler, ZKM, 1998/99\n\n",argv[0]);
  fprintf (stderr,"16bit, YUV version\n\n");
  
  /* Setup configuration filename */
  strcpy(cnffilename,argv[0]);
  strcat(cnffilename,".cnf");
  
  /* Set defaults, load existing configuration settings, generate table */
  DefaultSetup();
  LoadConfiguration();
  InitLookup();
  
  /* Set framegrabber device from command line */
  modeflag=0;
  inputflag=0;
  DevName = DefaultName;
  if (argc>1) {
   for (i=1; i<argc; i++) {
    if (argv[i][0]=='-') {
     if        (argv[i][1]=='p') {
      fprintf (stderr,"PAL set.\n");
      modeflag=0;
     } else if (argv[i][1]=='n') {
      fprintf (stderr,"NTSC set.\n");
      modeflag=1;
     } else if (argv[i][1]=='s') {
      fprintf (stderr,"SVIDEO set.\n");
      inputflag=0;
     } else if (argv[i][1]=='c') {
      fprintf (stderr,"COMPOSITE set.\n");
      inputflag=1;
     }
    } else {
     DevName=(char *)argv[i];
     fprintf (stderr,"DEVICE set to %s.\n",DevName);
    }
   }
  }
  
  /* Open framegrabber */   
  if ((the_grabber = open(DevName, O_RDONLY)) < 0) {
    perror("open failed");
    fprintf(stderr, "Is the module for device %s loaded?\n", DevName);
    exit(-1);
  }

  /* Setup framegrabber */
  geo.rows = ROWS;
  geo.columns = COLS;
  geo.frames = 1;
  geo.oformat = METEOR_GEO_RGB16;

  if (ioctl(the_grabber, METEORSETGEO, &geo) < 0) {
    perror("ioctl SetGeometry failed");
    exit(-1);
  }

  if (ioctl(the_grabber, METEORGFROFF, &off) < 0) {
    perror("ioctl GetFrameOffset failed");
    exit(-1);
  }

  if (modeflag) {
   fprintf(stderr, "Using NTSC format.");
   c = METEOR_FMT_NTSC;
  } else {
   fprintf(stderr, "Using PAL format.");
   c = METEOR_FMT_PAL;
  }
  if (ioctl(the_grabber, METEORSFMT, &c) < 0) {
    perror("ioctl SetFormat failed");
    exit(-1);
  }

  if (inputflag) {
   c = METEOR_INPUT_DEV0;
   fprintf(stderr, "Using Composite 0 input.");
  } else {
   c = METEOR_INPUT_DEV_SVIDEO;
   fprintf(stderr, "Using SVideo input.");
  }
  if (ioctl(the_grabber, METEORSINPUT, &c) < 0) {
    perror("ioctl Setinput failed");
    exit(-1);
  }
  
  /* Print processing settings */
  fprintf (stderr,"Current setting\n\n");
  fprintf (stderr,"Bounding-box-mode set to %i\n",boundmode);
  /* Set and print framegrabber controls */
  SetChcv();
  SetBrightness();
  SetChroma(); 
  SetContrast();
  SetHue();

  fprintf (stderr,"\n");

  /* Open window */
  setupDisplay();
  
  /* Map framebuffer memory */
  mmbuf=(char *)mmap((caddr_t)0, SIZE, PROT_READ,MAP_FILE|MAP_SHARED, the_grabber, (off_t)0);
  if ( !mmbuf ) {
    perror("mmap failed");
    exit(-1);
  }
  fbuf = mmbuf + off.frame_offset[0];

  /* Setup automatic grabbing */
  sigact.sa_handler= gotframe; 
  sigemptyset(&sigact.sa_mask);
  sigact.sa_flags= SA_RESTART;

  if ( sigaction(SIGUSR2, &sigact, NULL) ) {
    perror("sigaction failed");
    exit(-1);
  }

  c = SIGUSR2;
  if (ioctl(the_grabber, METEORSSIGNAL, &c) < 0) {
    perror("ioctl SetSignal failed");
    exit(-1);
  }

  /* Get semaphore */
  createsem(&semset_id);

  /* Create Network thread and initialize loop flag */
  in_loop=1;
  pthread_create (&thread, NULL, (void *)NetworkThread, NULL);
  
  /* Enable continuous frame capture */
  c = METEOR_CAP_CONTINOUS ;
  if (ioctl(the_grabber, METEORCAPTUR, &c)) {
    perror("ioctl CaptContinuous failed");
    GlobalExit();
  }
  
  /* Sit here in the loop and do nothing */
  while(in_loop) sleep(1);
  
}

/* X-windows setup */

static void setupDisplay() {
  int depth;
  int screen;
  XWindowAttributes win_attributes;
  XVisualInfo desired_visual_template;
  XVisualInfo *visual_info_list; 
  int no_visual_matched;
  int pixelNum;
  XGCValues gc_values;
  unsigned long event_mask;

  display_ = XOpenDisplay(NULL);
  wtwin_ = XCreateSimpleWindow(display_, DefaultRootWindow(display_),
			       0, 0, WINDOW_SIZE_X, WINDOW_SIZE_Y, 2,
			       BlackPixel(display_, DefaultScreen(display_)),
			       BlackPixel(display_, DefaultScreen(display_)));  
  XMapWindow(display_, wtwin_);
  XSelectInput(display_, wtwin_,KeyPressMask | KeyReleaseMask);
  gc_values.graphics_exposures = False;
  gc_ = XCreateGC(display_, wtwin_, GCGraphicsExposures, &gc_values);  

  screen = DefaultScreen(display_);
  XGetWindowAttributes(display_, wtwin_, &win_attributes);
  depth = win_attributes.depth;
  
  shminfo = (XShmSegmentInfo*) malloc (sizeof(XShmSegmentInfo));
  shmimage = XShmCreateImage(display_, DefaultVisual(display_, screen),
			     depth, ZPixmap, NULL, shminfo, WINDOW_SIZE_X,
			     WINDOW_SIZE_Y);
  shminfo->shmid = shmget(ftok(DevName, 'v'),
			  shmimage->bytes_per_line * shmimage->height,
			  IPC_CREAT | 0777);
  shminfo->shmaddr = (char *) shmat(shminfo->shmid, (void *) 0, 0);
  shmimage->data = shminfo->shmaddr;
  
  XShmAttach(display_, shminfo);
 
  desired_visual_template.screen = screen;
  desired_visual_template.class = TrueColor;
  visual_info_list = XGetVisualInfo(display_, 
				    VisualScreenMask | VisualClassMask,
				    &desired_visual_template,
				    &no_visual_matched);
  if (no_visual_matched == 0)
    fprintf(stderr, "No visual matched\n");
  
  event_mask = KeyPressMask | ButtonPressMask;
  XSelectInput(display_,wtwin_,event_mask);
   
}


